#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <gmp.h>
#include <unistd.h>
#include <sys/types.h>
#include "mpzqform.h"
#include "mpzsmoothscan.h"

/*
    Copyright 2009-2012 Andrew V. Sutherland

    This file is part of "smoothrelation".

    smoothrelation is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    smoothrelation is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with smalljac.  If not, see <http://www.gnu.org/licenses/>.
*/

#define VERSION			"1.3"

#define TEST_TIMING		0				// set to 1 to time form composition and smoothness testing spped

#define LOG2			0.69314718055994530941723212145818
#define log2(x)		(log(x)/LOG2)
#define delta_msecs(s,t)	((t-s)/(CLOCKS_PER_SEC/1000))		// assumes CLOCKS_PER_SEC is a multiple of 1000

#define MAX_B				50000			// maximum smoothness bound
#define MAX_BASE			5133			// set to pi(MAX_B), this is overkill
#define MAX_EBITS			(MAX_B/2)		// total across all exponents
#define MAX_K0			    6
#define MAX_RELATIONS		1				// maximum number of distinct relations to remember
#define BATCH_SIZE			2048
#define LOG2_BATCH_SIZE	    12
#define MAX_EXP_BITS		16
#define OMEGA			    1.25	    	// tuning parameter, see Bisson-Sutherland http://www.sciencedirect.com/science/article/pii/S0022314X09002789

/*
	The arrays below are statically allocated to keep them off the stack.
	These could easily be allocated dynamically.
*/
unsigned short r[MAX_RELATIONS][MAX_BASE];
mpz_t ga[MAX_EBITS], gb[MAX_EBITS];
int bitmap[MAX_EBITS];						// index of ell corresponding to each exp bit
char beta_e[MAX_EBITS];						// exponent bit-vector for the current beta
double wts[MAX_EBITS];						// binary logarithmic weight of each bit
mpz_t a[MAX_BASE],b[MAX_BASE];
unsigned long ell[2*MAX_BASE], ell0[MAX_BASE];
unsigned long h[MAX_BASE];
char hs[MAX_BASE];
int eoff[MAX_BASE], ebits[MAX_BASE];

struct exp_struct {
	unsigned short i[MAX_K0];              // index into factor base
	unsigned short e[MAX_K0];              // positive exponent
};

void smoothbox_init (mpz_t D, int B, mpz_t filter, double min_wt);
int smoothbox_feed (mpz_t out_a, mpz_t out_b, struct exp_struct out_exp[1], mpz_t a, mpz_t b, struct exp_struct in_exp[1]);
int smoothbox_flush (mpz_t out, mpz_t out_b, struct exp_struct out_exp[1]);

int divides_conductor (mpz_t D, mpz_t p)
{
	static mpz_t p2;
	static int init;
	
	if ( ! init ) { mpz_init(p2); init = 1; }
	mpz_mul(p2,p,p);
	if ( ! mpz_divisible_p(D,p2) ) return 0;
	if ( mpz_cmp_ui(p,2) != 0 ) return 1;
	return ( mpz_fdiv_ui(D,16) < 2 );
}


void evaluate_exp (mpz_t x, mpz_t y, struct exp_struct exp[1], int k0, mpz_t D, mpz_t L)
{
    static mpz_t u,v;
    static int init;
    int i;
    
    if ( ! init ) { mpz_init(u); mpz_init(v); init = 1; }
    mpz_qform_exp2 (x,y,a[exp->i[0]],b[exp->i[0]],exp->e[0],D,L);
    for ( i = 1 ; i < k0 ; i++ ) {
        mpz_qform_exp2 (u,v,a[exp->i[i]],b[exp->i[i]],exp->e[i],D,L);
        mpz_qform_compose2 (x,y,x,y,u,v,D,L,1);
    }
}


int main (int argc, char *argv[])
{
	clock_t start, end;
	gmp_randstate_t rstate;
	int s[MAX_K0], etot, etot2, esubtot;
	double maxNwt, wt, Bwt, sp_minwt;
	struct exp_struct in_exp, out_exp;
	mpz_t u,v,x,y,p,q,D,L,out_a, out_b,d,maxN, P, Pa, Pb, filter;
	unsigned long B, E,T, n, N, tests;
	register int i, j, k0, k, lastj, m, m0;
	int cnt, badcnt, btests, arity, i0;
	
	if ( argc < 3 ) {
		puts ("smoothrelation D B [P filter seed k0]");
		puts ("    D is an imaginary discriminant (if positive, it will be made negative)");
		puts ("    B is the smoothness bound");
		puts ("    p is the norm of an optional prime to be included in the relation (which may be greater than B).");
		puts ("    filter restricts the factor base by prohibiting primes whose norms divide filter.");
		puts ("    seed is a positive integer that determines a sequence of pseudo-random numbers.");
		puts ("    k0 is the number of primes that are allowed to occur with multiplicity > 1 (set to 4 by default, must lie in [2,6])");
		printf ("    software version %s\n", VERSION);
		return 0;
	}

	mpz_init(p);  mpz_init(q); mpz_init(D); mpz_init(L); mpz_init(d); mpz_init(u); mpz_init(v); mpz_init(x); mpz_init(y); mpz_init(maxN); mpz_init(out_a); mpz_init(out_b); mpz_init(P); mpz_init(Pa); mpz_init(Pb); mpz_init(filter);
	mpz_set_str(D,argv[1],0);
	if ( mpz_sgn(D)>0 ) mpz_neg(D,D);
	if ( ! mpz_qform_valid_discriminant(D) ) { gmp_printf ("D=%Zd is not a valid discriminant\n", D); return 0; }
	mpz_neg(L,D);
	mpz_fdiv_q_ui(maxN,L,3);								// bound on largest possible norm in cl(D), sqrt(-D/3)
	mpz_sqrt(maxN,maxN);  mpz_add_ui(maxN,maxN,1);	
	maxNwt = mpz_sizeinbase(maxN,2);						// upper bound on bits in largest possible norm
	mpz_root(L,L,4);
	B = atol (argv[2]);
	if ( argc > 3 ) {
		mpz_set_str(P,argv[3],0);
		if ( mpz_sgn(P) )
			if ( ! mpz_qform_primeform2(Pa,Pb,P,D) ) { gmp_printf ("Couldn't create primeform for P=%Zd\n", P); return 0; }
	}
	if ( argc > 4 ) mpz_set_str(filter,argv[4],0);
	if ( ! mpz_sgn(filter) ) mpz_set_ui(filter,1);
	gmp_randinit_default (rstate);
	if ( argc > 5 ) { 
		N = strtoul (argv[5], 0,0);
	} else {
		N = (((unsigned long)gethostid())<<32) + getpid();		// use a seed that depends on process id to ensure parallel jobs generate independent random forms
		N *= time(0);										// and throw in the current time for good measure
	}
	if ( argc > 6 ) k0 = atoi(argv[6]); else k0 = 4;
	if ( k0 < 3 || k0 > MAX_K0 ) { printf ("Invalid k0=%d, must be between 2 and %d\n", k0, MAX_K0); return 0; }
	E = 1UL<<40;											// we set E at an arbitrarily large value that we never plan to reach, relying on the user to pick a good B
	gmp_randseed_ui (rstate, N);
	printf ("Using seed %lu\n", N);
	for ( i = 0 ; i < MAX_BASE ; i++ ) { mpz_init(a[i]); mpz_init(b[i]); }

	// setup small primes minwt -- used to control early abort in smoothness testing
	// for large discriminants specify 0 to take the max default, but if |D| is small we don't want minwt too big
	sp_minwt = mpz_sizeinbase(D,2) / 16.0;
	if ( sp_minwt > 8.0 ) sp_minwt = 0.0;

	gmp_printf ("Searching for relation in cl(%Zd)\n", D);
	
	start = clock();
	tests = 0;
	cnt = 0;
	for (;;) {
		m = 0;
		mpz_set_ui(p,2);
		smoothbox_init(D,B,filter,sp_minwt);
		while ( m < MAX_BASE && mpz_cmp_ui(p,B) < 0 ) {
			if ( mpz_kronecker(D,p) > -1 && ! divides_conductor(D,p) && ! mpz_divisible_p(filter,p) ) {
				ell[m++] = mpz_get_ui(p);
				for ( mpz_set(q,p) ; mpz_cmp(q,maxN) <= 0 ; mpz_mul(q,q,p) );
			}
			mpz_nextprime(p,p);
		}
		printf ("Factor base has %d elements, largest norm is %ld\n", m, ell[m-1]);
		
		// Move norms of ideal classes with order 2 to the end of the list
		for ( i = k=0 ; i < m ; i++ ) {
			if ( mpz_divisible_ui_p(D,ell[i]) ) {
				mpz_set_ui(p,ell[i]);
				if ( ! mpz_qform_primeform2(x,y,p,D) ) { printf ("Fatal error, couldn't create primeform for ell=%ld\n",ell[i]); return 0; }
				mpz_set(v,y);
				mpz_qform_invert2(x,y,D);
				if ( mpz_cmp(v,y)==0 ) { ell[m+k++] = ell[i];  ell[i] = 0; }
			}
		}
		for ( i = j = 0 ; i < m+k ; i++ ) if ( ell[i] ) ell[j++] = ell[i];
		m0 = m-k;
		
		for ( i = 0 ; i < m ; i++ ) {
			mpz_set_ui(p,ell[i]);
			if ( ! mpz_qform_primeform2(a[i],b[i],p,D) ) { printf ("Fatal error, couldn't create primeform for ell=%ld\n",ell[i]); return 0; }
		}

#if TEST_TIMING
//		Timing code to test the speed of the group operation and smoothness testing
		mpz_set(u,a[m-1]); mpz_set(v,b[m-1]);
		nstart = clock();
		for ( i = 0 ; i < 1000000 ; i++ ) mpz_qform_compose2(u,v,u,v,a[m-2],b[m-2],D,L,1);
		end = clock();
		printf ("Time to compose 1000000 forms is %ld msecs (%ld/sec)\n", delta_msecs(nstart,end), 1000000000/(delta_msecs(nstart,end)+1));

		mpz_set(u,a[m-1]); mpz_set(v,b[m-1]);
		k = 0;
		nstart = clock();
		for ( i = 0 ; i < 1000000 ; i++ ) {
			mpz_qform_compose2(u,v,u,v,a[m-2],b[m-2],D,L,1);
			if ( smoothbox_feed(x,y,&out_exp,u,v,&in_exp) ) k++;
		}
		end = clock();
		printf ("Time to enum+test 1000000 forms for %ld-smoothness using Bernstein %ld msecs (%ld/sec) (found %d)\n", B, delta_msecs(nstart,end), 1000000000/(delta_msecs(nstart,end)+1), k);
#endif
		
		// compute exponent bounds based on B
//printf ("Exponent bounds for B=%d: ", B);
		Bwt = log2(B)+1;
		for ( i = j = 0, etot = 0 ; i < m0 ; i++ ) {
			wt = log2(ell[i]);
			ebits[i] = (int)ceil(OMEGA*(Bwt-wt));
			if ( ebits[i] > MAX_EXP_BITS ) ebits[i] = MAX_EXP_BITS;
			eoff[i] = etot;
//printf ("%d: %d bits at offset %d\n", ell[i],ebits[i], eoff[i]);
			if ( etot+ebits[i] > MAX_EBITS ) { printf ("MAX_EBITS=%d is too small, increase to at least %d\n", MAX_EBITS, etot); return 0; }
			mpz_init_set(ga[j],a[i]); mpz_init_set(gb[j],b[i]);
			bitmap[j] = i;
			wts[j]=wt;
			for ( j++ ; j < etot + ebits[i] ; j++ ) {
				mpz_qform_square2(ga[j],gb[j],ga[j-1],gb[j-1],D,L);
				bitmap[j] = i;
				wts[j] = (j-eoff[i])*wt;
//printf ("Bit %d has wt %.2f\n", j, wts[j]);
			}
			etot+=ebits[i];
		}
//printf("Exponent bits = %d, Max norm bits = %d\n", etot,(int)maxNwt);
		for ( i = etot2 = 0 ; i < m0 ; i++ ) { if ( ebits[i] < 2 ) break; etot2 += ebits[i]; }
		for ( N = 0 ; N < E ; N+=BATCH_SIZE ) {
			// pick starting point by setting k0 random bits in beta's exponent vector (making sure ell[s[i]] we use are distinct)
			memset(beta_e,0,etot);
			esubtot = 0;
			Bwt = 0.0;
			for ( i = 0 ; i < k0 ; i++ ) {
				do {
					j = gmp_urandomm_ui (rstate,etot2);
					s[i] = bitmap[j];
					for ( k = 0 ; k < i ; k++ ) if ( s[k]==s[i] ) break;
				} while ( k < i );
				esubtot += ebits[s[i]];
				Bwt += wts[j];
				beta_e[j] = 1;
				in_exp.e[i] = (1<<(j-eoff[s[i]]));
				in_exp.i[i] = s[i];
				if ( i ) {
					mpz_qform_compose2(u,v,u,v,ga[j],gb[j],D,L,1);
				} else {
					if ( ! mpz_sgn(Pa) ) {
						mpz_set(u,ga[j]); mpz_set(v,gb[j]);
					} else {
						mpz_qform_compose2(u,v,Pa,Pb,ga[j],gb[j],D,L,1);
					}
				}
			}
			// Make sure we have enough bits to work with
			if ( esubtot < LOG2_BATCH_SIZE ) { N-=BATCH_SIZE; continue; }
			badcnt = 0;
			lastj = -1;
			btests = 0;
			for ( n = 0 ; n < BATCH_SIZE ; n++ ) {
				if ( Bwt < maxNwt ) goto next_beta;
				badcnt = 0;
				tests++; btests++;
				if ( smoothbox_feed (out_a,out_b,&out_exp,u,v,&in_exp) ) {
//gmp_printf("Smooth form <%Zd,%Zd,*>\n", out_a, out_b);
//printf ("out exponent: "); for ( i = 0 ; i < k0 ; i++ ) if ( out_exp.e[i] ) printf ("<%ld,%d>, ", ell[out_exp.i[i]], out_exp.e[i]); puts ("");
					// now factor the norm with trial division, which is easy since it is smooth
					mpz_set(x,out_a);
					for ( i = 0 ; i < m ; i++ ) {
						for ( h[i] = 0 ; mpz_divisible_ui_p(x,ell[i]) ; h[i]++ ) mpz_divexact_ui(x,x,ell[i]);
						if ( h[i] ) hs[i] = ( mpz_fdiv_ui(out_b,2*ell[i]) > ell[i] ? -1 : 1 ); else hs[i] = 0;
					}
					// determine exponents for k0 primeforms; if these intersect the factorization of a, we need to add/sub, possibly changing signs
					// our sign convention will be to move the k0 forms to the other side, negating their signs (which were all nonnegative to begin with)
					for ( i0 = 0 ; i0 < k0 ; i0++ ) {
						i = out_exp.i[i0];  j = out_exp.e[i0];
						if ( h[i] ) {
							if ( hs[i] < 0 ) { h[i] += j; } else { if ( h[i] > j ) { h[i] -= j; } else { h[i] = j - h[i]; hs[i] = -1; } };
						} else {
							h[i] = j;  hs[i] = -1;
						}
					}
					// make sure relation is really new (only relevant when searching for multiple relations)
					for ( i = 0 ; i < cnt ; i++ ) {
						for ( j = 0 ; j < m ; j++ ) if ( h[j] != r[i][j] ) break;
						if ( j == m ) break;
					}
					if ( i == cnt ) {
						for ( j = 0, arity = 0 ; j < m ; j++ ) { r[i][j] = h[j]; if ( h[j] ) arity++; }
						if ( mpz_sgn(P) ) { 
							gmp_printf ("Smooth relation for %Zd: \n    ", P);
						} else {						
							printf ("Smooth %d-relation: ", arity);
						}
						for ( i = 0 ; i < m ; i++ ) if ( h[i] ) printf ("<%ld,%s%ld>, ", ell[i], (hs[i]<0?"-":""), h[i]);
						end = clock();
						cnt++;
						if ( cnt < MAX_RELATIONS ) {
							printf("    %ld msecs\n", delta_msecs(start,end));
							printf ("%ld tests total, %ld tests/s\n", tests, 1000UL*tests/(delta_msecs(start,end)+1));
						} else {
							puts("");
							goto done;
						}
					}
					break;
				}
next_beta:
				if ( badcnt++ > 10 ) break;
				do { j = gmp_urandomm_ui(rstate,esubtot); } while ( j==lastj);
				lastj = j;
				for ( i = 0 ; j >= ebits[s[i]] ; i++ )  j -= ebits[s[i]];
				j += eoff[s[i]];
				if ( beta_e[j] ) {
					if ( Bwt - wts[j] < maxNwt ) goto next_beta;
//printf ("clearing bit %d (%ld^%d)\n", j, ell[s[i]], (1<<(j-eoff[s[i]])));
					mpz_set(x,gb[j]);
					mpz_qform_invert2(ga[j],x,D);
					mpz_qform_compose2(u,v,u,v,ga[j],x,D,L,1);
					Bwt -= wts[j]; 
					in_exp.e[i] ^= (1<<(j-eoff[s[i]]));
				} else {
//printf ("setting bit %d (%ld^%d)\n", j, ell[s[i]], (1<<(j-eoff[s[i]])));
					mpz_qform_compose2(u,v,u,v,ga[j],gb[j],D,L,1);
					Bwt += wts[j];
					in_exp.e[i] |= (1<<(j-eoff[s[i]]));
				}
				beta_e[j]=1-beta_e[j];
				if ( Bwt < maxNwt ) goto next_beta;
			}
//printf ("%d of %d in batch tested\n", btests, BATCH_SIZE);
		}
		// currently we never get here because E is hard-wired to a ridiculously big value
		if ( 1.01*B < MAX_B ) {
			B = 1.05*B+1;
			E = 1.05*1.05*1.05*E+1;
			printf ("B=%ld, E=%ld\n", B, E);
		} else {
			printf ("At MAX_B=%d\n", MAX_B);
		}
	}

done:
	end = clock();
	T = delta_msecs(start,end);
	printf ("Performed a total of %ld tests, %ld tests/sec, total elapsed time %.3f secs\n", tests, (T?(1000UL*tests)/T:-1), (double)T/1000.0);
}


/*
	smoothbox should eventually be put into its own module
*/

#define SB_BS				4096		// smooth test batch size
#define SB_NSP			7			// maximum number of small primes for early abort

static mpz_t sb_P;
static mpz_t *sb_a, *sb_b, *sb_t;
static int sb_n, sb_nsp;
static int sb_sp[SB_NSP+1];
static double sb_spw[SB_NSP+1];
static struct exp_struct *sb_exp;
static double sb_min_wt;

void smoothbox_init (mpz_t D, int B, mpz_t filter, double min_wt)
{
	static mpz_t p,q;
	int i;
	
	if ( ! sb_a ) {
		mpz_init(p);  mpz_init(q);
		sb_exp = malloc(SB_BS*sizeof(struct exp_struct));
		sb_a = malloc(SB_BS*sizeof(mpz_t));
		for ( i = 0 ; i < SB_BS ; i++ ) mpz_init(sb_a[i]);
		sb_b = malloc(SB_BS*sizeof(mpz_t));
		for ( i = 0 ; i < SB_BS ; i++ ) mpz_init(sb_b[i]);
		sb_t = malloc(2*SB_BS*sizeof(mpz_t));
		for ( i = 0 ; i < 2*SB_BS ; i++ ) mpz_init(sb_t[i]);
		mpz_init(sb_P);
	}
	
	// don't bother being efficient about computing sb_P, we expect B to be fairly small
	mpz_set_ui(p,1);
	sb_nsp = 0;
	mpz_set_ui(sb_P,1);
	for (;;) {
		do { mpz_nextprime(p,p); } while ( mpz_kronecker(D,p) < 0 || divides_conductor(D,p) || mpz_divisible_p(filter,p) );		// AVS 9/30/12 changed to support filter (replaces unramified flag), and to handle 2 correctly
		if ( mpz_cmp_ui(p,B) > 0 ) break;
		if ( sb_nsp < SB_NSP ) { sb_sp[sb_nsp] = mpz_get_ui(p); sb_spw[sb_nsp] = log2(sb_sp[sb_nsp]);  sb_nsp++; }
		for ( mpz_set(q,p) ; mpz_cmp_ui(q,B*B) <= 0 ; mpz_mul(q,q,p), mpz_mul(sb_P,sb_P,p) );
	}
	if ( ! min_wt ) sb_min_wt = 1.2*SB_NSP; else sb_min_wt = min_wt;
	sb_n = 0;
}

int smoothbox_feed (mpz_t out_a, mpz_t out_b, struct exp_struct out_exp[1] , mpz_t a, mpz_t b, struct exp_struct in_exp[1])
{
	register double wt;
	register unsigned long q;
	register int i;

	i = ( sb_nsp && sb_sp[0] == 2 ? 1 : 0 );
	wt = ( i ? mpz_scan1(a,0) : 0.0);
	for ( ; i < sb_nsp ; i++ )
		for ( q = sb_sp[i] ; mpz_divisible_ui_p(a,q) ; q *= sb_sp[i] ) wt += sb_spw[i];
	if ( wt >= sb_min_wt ) {
		mpz_set(sb_a[sb_n],a);  mpz_set(sb_b[sb_n],b); sb_exp[sb_n]=in_exp[0]; sb_n++;
		if ( sb_n==SB_BS ) return smoothbox_flush(out_a, out_b, out_exp);
	}
	return 0;
}

int smoothbox_flush (mpz_t out_a, mpz_t out_b, struct exp_struct out_exp[1])
{
	int i;

	for ( i = sb_n ; i < SB_BS ; i++ ) mpz_set_ui(sb_a[i],1);
	i = mpz_smoothscan (sb_a, SB_BS, 0, sb_P, sb_t);
	if ( i < sb_n ) {
		mpz_set(out_a, sb_a[i]);  mpz_set(out_b, sb_b[i]);  out_exp[0] = sb_exp[i];
		sb_n = 0; return 1;
	}
	sb_n = 0;  return 0;
}

